//
// TSIPluginInterfaceInternal.h
// 
// General header file for TBV plugin support
//

#ifndef TSIPLUGININTERFACE_H
#define TSIPLUGININTERFACE_H

#include "global.h" // to get platform (OS) definitions (plus macros, qxRound())

// make sure that "MAKE_PLUGIN" is defined here or as a preprocessor definition (this is only important on Windows)
//
#define MAKE_PLUGIN

#if defined OS_WIN32
#ifdef MAKE_PLUGIN
#define PLUGIN_ACCESS __declspec(dllexport)
#else
#define PLUGIN_ACCESS __declspec(dllimport)
#endif
#else
// for all other platforms PLUGIN_ACCESS is empty (not needed)
#define PLUGIN_ACCESS
#endif


#if defined(OS_WIN32)
#include <windows.h>
#include <malloc.h>
#endif
#include <stdio.h>
#include <stdlib.h>

#ifndef OS_WIN32
#include <dlfcn.h>
#endif

#ifdef OS_WIN32
#define GETSYMADDRESS GetProcAddress
#else
#define GETSYMADDRESS dlsym
#endif

#include <inttypes.h>

class PLUGIN_ACCESS TSIPluginInterface
{
public:	 
	TSIPluginInterface() {};
	virtual ~TSIPluginInterface() {};
	
	// this function will be called once
	//
	virtual bool initPlugin() = 0;
	
	// these methods will be called to execute the plugin (your essential code)
	// the "executePre/PostRun()" functions will be called once while the
	// executeTimePoint()" function will be called at each incremental processing step
	//
	virtual bool executePreRun(uint16_t port = 0) = 0;
	virtual bool executeTimePoint() = 0;
	virtual bool executePostRun() = 0;
};


//
// These are typedef's for dynamic access of TSI (Turbo-Satori) functions - version with "static" definitions
// Commands start with "t" to indicate that these are Turbo-Satori access functions

// general functions
typedef void (*GetVersionOfTSIPrototype)(int *, int *, int *);
static GetVersionOfTSIPrototype tGetVersionOfTSI;

typedef void (*GetTSIPathPrototype)(char *);
static GetTSIPathPrototype tGetTSIPath;

typedef void (*GetPluginsPathPrototype)(char *);
static GetPluginsPathPrototype tGetPluginsPath;


// output in log window
typedef void (*LogTextPrototype)(const char *);
static LogTextPrototype tLogText;

// dialogs to get/provide information - note these fns are only usable in non-incremental calls (ignored in "execute()" plugin fn)
typedef void (*MessageBoxPrototype)(const char *);
static MessageBoxPrototype tMessageBox;

typedef int (*YesNoMessageBoxPrototype)(const char *);
static YesNoMessageBoxPrototype tYesNoMessageBox;

typedef int (*GetIntegerInputPrototype)(const char *, int, int, int);
static GetIntegerInputPrototype tGetIntegerInput;

typedef float (*GetFloatInputPrototype)(const char *, float, float, float);
static GetFloatInputPrototype tGetFloatInput;

typedef int (*GetTextInputPrototype)(const char *, const char *, char *);
static GetTextInputPrototype tGetTextInput;

typedef int (*GetOpenFileNamePrototype)(const char *, const char *, const char *, char *);
static GetOpenFileNamePrototype tGetOpenFileName;

typedef int (*GetSaveFileNamePrototype)(const char *, const char *, const char *, char *);
static GetSaveFileNamePrototype tGetSaveFileName;


// TSI info
typedef int (*GetCurrentTimePointPrototype)();
static GetCurrentTimePointPrototype tGetCurrentTimePoint;

typedef int (*GetNrOfChannelsPrototype)();
static GetNrOfChannelsPrototype tGetNrOfChannels;

typedef void (*GetValuesFeedbackFolderPrototype)(char *);
static GetValuesFeedbackFolderPrototype tGetValuesFeedbackFolder;

typedef void (*GetImagesFeedbackFolderPrototype)(char *);
static GetImagesFeedbackFolderPrototype tGetImagesFeedbackFolder;

// Selected channel data
typedef int (*GetNrOfSelectedChannelsPrototype)();
static GetNrOfSelectedChannelsPrototype tGetNrOfSelectedChannels;

typedef int *(*GetSelectedChannelsPrototype)();
static GetSelectedChannelsPrototype tGetSelectedChannels;

// Get raw data values
typedef float (*GetRawDataScaleFactorPrototype)();
static GetRawDataScaleFactorPrototype tGetRawDataScaleFactor;

typedef float (*GetRawDataWL1Prototype)(int, int);
static GetRawDataWL1Prototype tGetRawDataWL1;

typedef float (*GetRawDataWL2Prototype)(int, int);
static GetRawDataWL2Prototype tGetRawDataWL2;

// Get preprocessed oxy/deoxy data (or preprocessed raw data if not oxy converted)
typedef int (*IsDataOxyDeoxyConvertedPrototype)();
static IsDataOxyDeoxyConvertedPrototype tIsDataOxyDeoxyConverted;

typedef int(*GetOxyDeOxyBaselineEndPrototype)();
static GetOxyDeOxyBaselineEndPrototype tGetOxyDeOxyBaselineEnd;

typedef float (*GetOxyDataScaleFactorPrototype)();
static GetOxyDataScaleFactorPrototype tGetOxyDataScaleFactor;

typedef float (*GetDataOxyPrototype)(int, int);
static GetDataOxyPrototype tGetDataOxy;

typedef float (*GetDataDeoxyPrototype)(int, int);
static GetDataDeoxyPrototype tGetDataDeoxy;



// Get preprocessed era pointers 
typedef int(*GetERAWindowPrototype)();
static GetERAWindowPrototype tGetERAWindow;

typedef double(*GetDataERAOxyPrototype)(int, int, int);
static GetDataERAOxyPrototype tGetERAOxy;

typedef double(*GetDataERADeOxyPrototype)(int, int, int);
static GetDataERADeOxyPrototype tGetERADeOxy;

typedef double(*GetDataERATotalPrototype)(int, int, int);
static GetDataERATotalPrototype tGetERATotal;

typedef double(*GetDataERASTDOxyPrototype)(int, int, int);
static GetDataERASTDOxyPrototype tGetERASTDOxy;

typedef double(*GetDataERASTDDeOxyPrototype)(int, int, int);
static GetDataERASTDDeOxyPrototype tGetERASTDDeOxy;

typedef double(*GetDataERASTDTotalPrototype)(int, int, int);
static GetDataERASTDTotalPrototype tGetERASTDTotal;




//Protocol, DM, GLM Functions
typedef int(*GetFullNrOfPredictorsPrototype)();
static GetFullNrOfPredictorsPrototype tGetFullNrOfPredictors;

typedef int(*GetNrOfPredictorsPrototype)();
static GetNrOfPredictorsPrototype tGetNrOfPredictors;

typedef int(*GetNrOfConfoundPredictorsPrototype)();
static GetNrOfConfoundPredictorsPrototype tGetNrOfConfoundPredictors;

typedef float(*GetValueOfDesignMatrixPrototype)(int, int, bool);
static GetValueOfDesignMatrixPrototype tGetValueOfDesignMatrix;

typedef float(*GetBetaOfChannelPrototype)(int, int, bool);
static GetBetaOfChannelPrototype tGetBetaOfChannel;

typedef float(*GettValueOfChannelPrototype)(int, bool, int, int *);
static GettValueOfChannelPrototype tGettValueOfChannel;

typedef int(*GetProtocolConditionPrototype)(int);
static GetProtocolConditionPrototype tGetProtocolCondition;

//Basic project functions
typedef float(*GetSamplingRatePrototype)();
static GetSamplingRatePrototype tGetSamplingRate;

//SVM access functions
typedef int(*tGetNumberOfClassesPrototype)();
static tGetNumberOfClassesPrototype tGetNumberOfClasses;

typedef float(*tGetCurrentClassifierOutputPrototype)();
static tGetCurrentClassifierOutputPrototype tGetCurrentClassifierOutput;


// Locally defined functions
//
inline bool InitTSIAccess()
{
#ifdef OS_WIN32

	HMODULE hndl = GetModuleHandleA("Turbo-Satori.exe");

	if(hndl == 0)
		hndl = GetModuleHandleA("Turbo-Satori_SimulMode_4WL.exe");

	if (hndl == 0)
		hndl = GetModuleHandleA("Turbo-Satori_SimulMode.exe");

	if(hndl == 0)
		return false;

#else

	// handle Mac and Linux
	
	void *hndl = dlopen(NULL, RTLD_LAZY);  // NULL pointer to get handle to main process
	if(hndl == NULL)
		return false;

#endif

	//
	// general functions
	//
	tGetVersionOfTSI = (GetVersionOfTSIPrototype)GETSYMADDRESS( hndl, "tGetVersionOfTSI" );
	if(tGetVersionOfTSI == NULL) return false;

	// IMPORTANT: If we do not get a pointer, we do not return false
	//  To use specific commands, a) you must check the obtained version string and use only those commands available in that version
	//                         or b) Check whether the function pointers are non-NULL for each command you want to use in your code

	tGetTSIPath = (GetTSIPathPrototype)GETSYMADDRESS( hndl, "tGetTSIPath" );
	tGetPluginsPath = (GetPluginsPathPrototype)GETSYMADDRESS( hndl, "tGetPluginsPath" );

	// input and display (GUI) functions
	//
	tLogText = (LogTextPrototype)GETSYMADDRESS( hndl, "tLogText" );
	// dialogs can only be used outside "execute()" fn, e.g. in "executePreRun()"
	tMessageBox = (MessageBoxPrototype)GETSYMADDRESS( hndl, "tMessageBox" );
	tYesNoMessageBox = (YesNoMessageBoxPrototype)GETSYMADDRESS( hndl, "tYesNoMessageBox" );
	tGetIntegerInput = (GetIntegerInputPrototype)GETSYMADDRESS( hndl, "tGetIntegerInput" );
	tGetFloatInput = (GetFloatInputPrototype)GETSYMADDRESS( hndl, "tGetFloatInput" );
	tGetTextInput = (GetTextInputPrototype)GETSYMADDRESS( hndl, "tGetTextInput" );
	tGetOpenFileName = (GetOpenFileNamePrototype)GETSYMADDRESS( hndl, "tGetOpenFileName" );
	tGetSaveFileName = (GetSaveFileNamePrototype)GETSYMADDRESS( hndl, "tGetSaveFileName" );
	
	// Basic TBV access functions
	//
	tGetCurrentTimePoint = (GetCurrentTimePointPrototype)GETSYMADDRESS( hndl, "tGetCurrentTimePoint" );
//	tGetExpectedNrOfTimePoints = (GetExpectedNrOfTimePointsPrototype)GETSYMADDRESS( hndl, "tGetExpectedNrOfTimePoints" );
	tGetNrOfChannels = (GetNrOfChannelsPrototype)GETSYMADDRESS( hndl, "tGetNrOfChannels" );
	tGetValuesFeedbackFolder = (GetValuesFeedbackFolderPrototype)GETSYMADDRESS( hndl, "tGetValuesFeedbackFolder" );
	tGetImagesFeedbackFolder = (GetImagesFeedbackFolderPrototype)GETSYMADDRESS( hndl, "tGetImagesFeedbackFolder" );

	// Selected channels
	//
	tGetNrOfSelectedChannels = (GetNrOfSelectedChannelsPrototype)GETSYMADDRESS( hndl, "tGetNrOfSelectedChannels" );
	tGetSelectedChannels = (GetSelectedChannelsPrototype)GETSYMADDRESS( hndl, "tGetSelectedChannels" );
	
	// Raw data
	//
	tGetRawDataScaleFactor = (GetRawDataScaleFactorPrototype)GETSYMADDRESS( hndl, "tGetRawDataScaleFactor" );
	tGetRawDataWL1 = (GetRawDataWL1Prototype)GETSYMADDRESS( hndl, "tGetRawDataWL1" );
	tGetRawDataWL2 = (GetRawDataWL2Prototype)GETSYMADDRESS( hndl, "tGetRawDataWL2" );

	// Preprocessed raw data or (most commonly) preprocessed oxy/deoxy data
	tIsDataOxyDeoxyConverted = (IsDataOxyDeoxyConvertedPrototype)GETSYMADDRESS( hndl, "tIsDataOxyDeoxyConverted" );
	tGetOxyDeOxyBaselineEnd = (GetOxyDeOxyBaselineEndPrototype)GETSYMADDRESS(hndl, "tGetOxyDeOxyBaselineEnd");
	tGetOxyDataScaleFactor = (GetOxyDataScaleFactorPrototype)GETSYMADDRESS( hndl, "tGetOxyDataScaleFactor" );
	tGetDataOxy = (GetDataOxyPrototype)GETSYMADDRESS( hndl, "tGetDataOxy" );
	tGetDataDeoxy = (GetDataDeoxyPrototype)GETSYMADDRESS( hndl, "tGetDataDeoxy" );

	// Get preprocessed era pointers 
	tGetERAWindow = (GetERAWindowPrototype)GETSYMADDRESS(hndl, "tGetERAWindow");
	tGetERAOxy = (GetDataERAOxyPrototype)GETSYMADDRESS(hndl, "tGetERAOxy");
	tGetERADeOxy = (GetDataERADeOxyPrototype)GETSYMADDRESS(hndl, "tGetERADeOxy");
	tGetERATotal = (GetDataERATotalPrototype)GETSYMADDRESS(hndl, "tGetERATotal");
	tGetERASTDOxy = (GetDataERASTDOxyPrototype)GETSYMADDRESS(hndl, "tGetERASTDOxy");
	tGetERASTDDeOxy = (GetDataERASTDDeOxyPrototype)GETSYMADDRESS(hndl, "tGetERASTDDeOxy");
	tGetERASTDTotal = (GetDataERASTDTotalPrototype)GETSYMADDRESS(hndl, "tGetERASTDTotal");

	tGetFullNrOfPredictors = (GetFullNrOfPredictorsPrototype)GETSYMADDRESS(hndl, "tGetFullNrOfPredictors");
	tGetNrOfPredictors = (GetFullNrOfPredictorsPrototype)GETSYMADDRESS(hndl, "tGetNrOfPredictors");
	tGetNrOfConfoundPredictors = (GetFullNrOfPredictorsPrototype)GETSYMADDRESS(hndl, "tGetNrOfConfoundPredictors");
	tGetValueOfDesignMatrix = (GetValueOfDesignMatrixPrototype)GETSYMADDRESS(hndl, "tGetValueOfDesignMatrix");
	tGetBetaOfChannel = (GetBetaOfChannelPrototype)GETSYMADDRESS(hndl, "tGetBetaOfChannel");
	tGettValueOfChannel = (GettValueOfChannelPrototype)GETSYMADDRESS(hndl, "tGettValueOfChannel");
	tGetProtocolCondition = (GetProtocolConditionPrototype)GETSYMADDRESS(hndl, "tGetProtocolCondition");

	//Basic project functions
	tGetSamplingRate = (GetSamplingRatePrototype)GETSYMADDRESS(hndl, "tGetSamplingRate");

	//SVM access functions
	tGetNumberOfClasses = (tGetNumberOfClassesPrototype)GETSYMADDRESS(hndl, "tGetNumberOfClasses");
	tGetCurrentClassifierOutput = (tGetCurrentClassifierOutputPrototype)GETSYMADDRESS(hndl, "tGetCurrentClassifierOutput");


	return true;
}

inline bool checkVersionGreaterEqualVersion(int MajorA, int MinorA, int SubMinorA, int MajorB, int MinorB, int SubMinorB)
{
	if(MajorA > MajorB)
		return true;

	if(MajorB > MajorA)
		return false;

	// Major is equal, Minor decides

	if(MinorA > MinorB)
		return true;

	if(MinorB > MinorA)
		return false;

	// Minor is equal, SubMinor decides

	if(SubMinorA >= SubMinorB)
		return true;

	// B smaller
	return false;
}


#endif // TSIPLUGININTERFACE_H
